//
//  Action Replay FileServer
//
//  (C)1997,by Craig Graham
//

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <io.h>
#include <fcntl.h>
#include <setjmp.h>
#include "ansicons.h"

// Defaults
//  maximum number of files open at once
#define MAX_SERV_HANDLES  10
//  io port for the action replay card - can be set at runtime with the /I switch, this is just a default
#define DEFAULT_IOPORT    0x320

// IO port access for Win95/DOS using Visual C++
#define inp(x)   ((unsigned char)_inp((unsigned short)(x)))
#define outp _outp

typedef unsigned char	uchar;
typedef unsigned short	ushort;
typedef unsigned long	ulong;

unsigned short io_port=0x320;    // 0x300

short quiet_mode=0,connected=0,stat=0;
jmp_buf recovery_jump;

unsigned char *fbuffer;   // pointer to a buffer used for file io
FILE *serv_handles[MAX_SERV_HANDLES+1]; // file handles we have available/used
long last_file_pos[MAX_SERV_HANDLES+1]; // used for retry on a read

short activity=0,act_t=1;
char *status_chars="SDLOCR";

void DisplayStatusLine(void)
{
  SetTextColour(CON_FYELLOW);
  SetTextColour(CON_BBLUE);
  printf("\033[s");
  SetTextXY(0,0);
  printf("AR FileServer For Playstation | IOPORT=%3x | quiet=",io_port);
  if(quiet_mode)
    printf("ON ");
  else
    printf("OFF");

  if(connected)
    printf(" | online");
  else
    printf(" |       ");

  printf(" |  status:%c     ",status_chars[stat]);

  SetTextColour(CON_FWHITE);
  SetTextColour(CON_BBLACK);
  printf("\033[u");
}

void BaseCheckForEsape(void)
{
  char key;

  if (kbhit())
  {
    key=getch();

    if(key=0x1B)
    {
 	    printf("ABORTED!\n");
      exit(-1);
    }
  }
}

void ResetPSX(void)
{
  char data;
//  outp(io_port, 0x69);
  outp(io_port, 0x69);
  while(inp(io_port+2) & 0x01)
    BaseCheckForEsape();
  data=inp(io_port);
  outp(io_port, 0x6E);
  while(inp(io_port+2) & 0x01)
    BaseCheckForEsape();
  data=inp(io_port);
  outp(io_port, 0x14);
  while(inp(io_port+2) & 0x01)
    BaseCheckForEsape();
  data=inp(io_port);
  outp(io_port, 0xbf);
  while(inp(io_port+2) & 0x01)
    BaseCheckForEsape();
  data=inp(io_port);
  outp(io_port, 0xc0);
  while(inp(io_port+2) & 0x01)
    BaseCheckForEsape();
  data=inp(io_port);
  outp(io_port, 0);
  while(inp(io_port+2) & 0x01)
    BaseCheckForEsape();
  data=inp(io_port);
  outp(io_port, 0);
  while(inp(io_port+2) & 0x01)
    BaseCheckForEsape();
  data=inp(io_port);
  outp(io_port, 0);

  connected=0;
}
  
void CheckForEscape(void)
{
  char key;

  if (kbhit())
  {
    key=getch();

    switch(key)
    {
      case 0x1B:
  	    printf("ABORTED!\n");
	      exit(-1);
        break;

      case 'r':
      case 'R':
        SetTextColour(CON_FRED);
        printf("**REBOOT\n");
        DisplayStatusLine();
        ResetPSX();
        longjmp(recovery_jump,'r');
        break;
    }
	}
}

unsigned char read_byte(void)
{
  if((act_t)&&(((activity++)&0x1ff)==0))
  {
    printf("\033[s");
    SetTextXY(77,0);
    SetTextColour(CON_FGREEN);
    SetTextColour(CON_BBLUE);
    if(activity&0x200)
    {
      printf("I");
    }else{
      printf(" ");
    }
    printf("\033[u");
    SetTextColour(CON_FWHITE);
    SetTextColour(CON_BBLACK);
  }

  while(inp(io_port+2) & 0x01)
    CheckForEscape();
  return inp(io_port);
}

// Send long
void ServSendLong(long v)
{
  long temp;
  char data;

  temp=v>>24;
  outp(io_port, temp&0xff);
  data=read_byte();
  temp=v>>16;
  outp(io_port, temp&0xff);
  data=read_byte();
  temp=v>>8;
  outp(io_port, temp&0xff);
  data=read_byte();
  outp(io_port, v&0xff);
  data=read_byte();
}

void diags(void)
{
  char data;
  
  // Send ACK for printing
  outp(io_port, 'p');

  // Skip first byte
  data=read_byte();

  SetTextColour(CON_FWHITE);

  // Get text string
  do{
    outp(io_port, 'p');
    data=read_byte();

    if(data)
      printf("%c",data);

  }while(data);

  DisplayStatusLine();
}

void serv_close_all_files(void)
{
  short f;
  for(f=0; f<MAX_SERV_HANDLES; f++)
  {
    if(serv_handles[f])
      fclose(serv_handles[f]);
  }
}

// Block file load
void serv_file(void)
{
  unsigned char data=1,fname[200],*ib;
  FILE *i;
  long ifsize,temp;

  // Send ACK for load op
  outp(io_port, 'l');

  // Skip first byte
  data=read_byte();

  // Get filename
  for(ib=&fname[0]; data!=0xff; ib++)
  {
    outp(io_port, 'l');

    data=read_byte();
    *ib=data;

  }

  *ib='\0';

  if(!quiet_mode)
  {
    SetTextColour(CON_FGREEN);
    printf("\n**SERV_FILE[%s]\n",fname);
    DisplayStatusLine();
  }

  // Open file
  i=fopen(fname,"rb");
  if(!i)
  {
    SetTextColour(CON_FRED);
    printf("**ERROR, FILE NOT FOUND\n");
    DisplayStatusLine();
    outp(io_port, 'n');

    data=read_byte();
    return;
  }

  // Send file found ack
  outp(io_port, 'f');
  data=read_byte();


  // Get file size
  fseek(i,0L,2);
  ifsize=ftell(i);    
  fseek(i,0L,0);  

  if(!quiet_mode)
  {
    SetTextColour(CON_FGREEN);
    printf("**FILE SIZE=%d=0x%x\n",ifsize,ifsize);
    DisplayStatusLine();
  }

  // Send file size to psx  
  ServSendLong(ifsize);

  // Read file into buffer
  fread((void*)fbuffer, 1, ifsize, i);

  // Send file out to PSX
  for(temp=0; temp<ifsize; temp++)
  {
    outp(io_port, fbuffer[temp]);
    data=read_byte();  
  }

  fclose(i);
}

// Remote file open
void serv_open(void)
{
  unsigned char data=1,fname[200],*ib;
  char file_handle_index;

  for(file_handle_index=0; (serv_handles[file_handle_index]!=NULL)&&(file_handle_index<MAX_SERV_HANDLES); file_handle_index++);
  
  // Send ACK for open op
  outp(io_port, 'o');

  // Skip first byte
  data=read_byte();

  // Get filename
  ib=&fname[0];
  do{
    outp(io_port, 'o');
    data=read_byte();
    *ib++=data;
  }while(data!=0xff);

  ib[-1]='\0';

  // Open file
  if(file_handle_index<MAX_SERV_HANDLES)
  {
    
    serv_handles[file_handle_index]=fopen(fname,"rb");
    if(!serv_handles[file_handle_index])
    {
      SetTextColour(CON_FRED);
      printf("**ERROR, FILE [%s] NOT FOUND\n",fname);
      DisplayStatusLine();
      file_handle_index=-1;
    }else{
      file_handle_index+=5;
      if(!quiet_mode)
      {
        SetTextColour(CON_FGREEN);
        printf("\n**SERV_OPEN[%s]:handle=%d\n",fname,file_handle_index);
        DisplayStatusLine();
      }
    }

  }else{
    SetTextColour(CON_FRED);
    printf("**ERROR:SERV_OPEN, TO MANY FILES OPEN\n");
    DisplayStatusLine();
    file_handle_index=-1;
  }

  // Send file handle
  outp(io_port, file_handle_index);
  data=read_byte();

}

// Close a remote file handle
void serv_close(void)
{
  char file_handle_index;

  // Send ACK for close op
  outp(io_port, 'c');

  // Skip first byte
  file_handle_index=read_byte();

  // Get file handle
  outp(io_port, 'c');
  file_handle_index=read_byte();

  if(!quiet_mode)
  {
    SetTextColour(CON_FGREEN);
    printf("**SERV_CLOSE:handle=%d\n",file_handle_index);
    DisplayStatusLine();
  }

  file_handle_index-=5;

  // Close the file if it was open
  if(serv_handles[file_handle_index])
  {
    fclose(serv_handles[file_handle_index]);
    serv_handles[file_handle_index]=NULL;
  }

}

void serv_unread_last(void)
{
  int file_handle_index;

  outp(io_port, 'x');
  read_byte();

  outp(io_port, 'x');

  file_handle_index=read_byte();
  file_handle_index-=5;

  fseek(serv_handles[file_handle_index],last_file_pos[file_handle_index],0);

  SetTextColour(CON_FYELLOW);
  printf("**WARNING:comms failure, re-reading:h=%d, from pos=%ld\n",file_handle_index+5,last_file_pos[file_handle_index]);
}

void serv_read(void)
{
  unsigned char data=1;
  int file_handle_index;
  long temp,read_len;

  // Send ACK for read op
  outp(io_port, 'z');

  // Skip first byte
  data=read_byte();

  // Get file handle
  outp(io_port, 'o');
  file_handle_index=read_byte();

  file_handle_index-=5;

  if((file_handle_index>MAX_SERV_HANDLES)||(serv_handles[file_handle_index]==NULL))
  {
    SetTextColour(CON_FRED);
    printf("**SERV_READ:INVALID FILE HANDLE[%d]\n");
    DisplayStatusLine();
    ServSendLong(-1);
    return;
  }

  // Get read length
  outp(io_port, 'z');
  temp=read_byte();
  read_len=temp<<24;
  outp(io_port, 'z');
  temp=read_byte();
  read_len|=temp<<16;
  outp(io_port, 'z');
  temp=read_byte();
  read_len|=temp<<8;
  outp(io_port, 'z');
  temp=read_byte();
  read_len|=temp;

  last_file_pos[file_handle_index]=ftell(serv_handles[file_handle_index]);

  // Read file into buffer
  temp=read_len;
  read_len=fread((void*)fbuffer, 1, read_len, serv_handles[file_handle_index]);

  if(!quiet_mode)
  {
    SetTextColour(CON_FGREEN);
    printf("**SERV_READ:h=%d, rql=%ld rdl=%ld\n",file_handle_index+5,temp,read_len);
    DisplayStatusLine();
  }

  // Send read size to psx  
  ServSendLong(read_len);

  data=0xff;
  // Send file data out to PSX
  for(temp=0; (temp<read_len)&&(data==0xff); temp++)
  {
    outp(io_port, fbuffer[temp]);
    data=read_byte();
  }

  if(data!=0xff)
  {
    fseek(serv_handles[file_handle_index],last_file_pos[file_handle_index],0);
    SetTextColour(CON_FYELLOW);
    printf("**WARNING:comms failure[%x], read failled:h=%d, returned to pos=%ld\n",data,file_handle_index+5,last_file_pos[file_handle_index]);
  }

}
  
char *psx_string="**ars_resync";
char *pc_string ="!!OK-LETS GO";

void do_handshake(void)
{
  char data='*';
  short i;

  outp(io_port, '!');

  for(i=0; (data==psx_string[i])&&(i<strlen(pc_string)); i++)
  {
    data=read_byte();
  }
}

// Console / File Server mode
void Server(void)
{
  unsigned char cmd=0;
  long f;

  setjmp(recovery_jump);

  printf("============\n");
  printf("Ready...\n");

  // zero all remote access file handles
  for(f=0; f<MAX_SERV_HANDLES; f++)
  {
    serv_handles[f]=NULL;
  }

  do{
     stat=0;
     DisplayStatusLine();

     outp(io_port, 0);

     // Wait for a request from the playstation
     cmd=read_byte();

     /// Playstation command parser
     switch(cmd)
     {
       case 'P': // print
         stat=1;
         diags();
         break;
     
       case 'L': // load whole file
         stat=2;
         serv_file();
         break;

       case 'R': // reconnect
         SetTextColour(CON_BGREEN);
         SetTextColour(CON_FWHITE);
         printf("                   PLAYSTATION CONNECT (Yaroze+CodeWarrior)                    ");
         SetTextColour(CON_BBLACK);
         printf("\n");
         serv_close_all_files();       // in case of a crash/reboot with files open, close them all
         connected=1;
         break;

      case 'O': // open file
        stat=3;
        serv_open();
        break;

      case 'C': // close file
        stat=4;
        serv_close();
        break;

      case 'Z': // read n bytes from file
        stat=5;
        serv_read();
        break;

      case 'X': // read comms failure, retry from previous read position
        serv_unread_last();
        break;

      case 'p': // poll for response
        do_handshake();
        break;

      case 0xff:  // error
      default:
        printf("**getting error [%x]!!!!\n",cmd);
        printf("**resyncing....\n");
        do{
          for(f=0; f<0x20000; f++)    // kill time to force a link timeout
              cmd++;
          outp(io_port, 0);
        }while(read_byte()==cmd);
        break;

      case 0:     // Bootup
        printf("PSX-BOOT\n");
        break;
    }

  }while(1);
}

void instructions(void)
{
  printf("Usage: ARS [/I<n>] [/Q] [/N]\n");
  printf("       /I<n>  use IO port <n> for Action Replay card (default=0x%x)\n",DEFAULT_IOPORT);
  printf("       /Q     quiet mode - supress all none-error messages in FileServer mode\n");
  printf("       /N     disable the link activity indicator (slightly better performance)\n");
}

void main(int argc, char *argv[])
{
  char *fname;

  SetTextMode(CON_MODE80x25);

  ClearScreen();
  DisplayStatusLine();
  SetTextColour(CON_BGREEN);
  SetTextColour(CON_FRED);
  printf("\nAction Replay FileServer");
  SetTextColour(CON_BBLACK);
  printf("\n");
  DisplayStatusLine();
  printf("Version:%s\n",__DATE__);
  DisplayStatusLine();
  printf("(C)1997,by Craig Graham\n");
  DisplayStatusLine();

  fbuffer=(unsigned char *)malloc(500000);

  // check for command line parameters
  while (--argc)
  {
  	if (argv[argc][0] != '/')
    {
      fname = argv[argc];
    }else{

      switch (argv[argc][1])
	    {
        case 'Q':     // Quite mode
        case 'q':
          quiet_mode=1;
          printf("Quiet mode is ON\n");
          break;

        case 'N':     // No activity indication
          act_t=0;
          printf("Activity indicator is OFF\n");
          break;

        case 'I':     // Set IO port
        case 'i':
          sscanf(argv[argc] + (argv[argc][2] == ':' ? 3 : 2), "%X",&io_port);
          break;

        default:
          instructions();
          exit(-1);
          break;

     }
    }
  }
    
  Server();

  exit(0);
}
